import csv
import lmdb
import numpy as np
import pyarrow as pa
import pickle as pk
import torch
from torch.utils.data import Dataset
import torchvision.transforms as transforms
from skimage.transform import resize
from bigearthnet_common.constants import BAND_STATS
from bigearthnet_common.base import ben_19_labels_to_multi_hot
import torchvision.transforms
import torchvision.transforms as transforms


def interp_band(bands, img10_shape=[120, 120]):

    bands_interp = np.zeros([bands.shape[0]] + img10_shape).astype(np.float32)
    for i in range(bands.shape[0]):
        bands_interp[i] = resize(bands[i] / 30000, img10_shape, mode='reflect') * 30000

    return bands_interp




def loads_pyarrow(buf):

 
    return pk.loads(buf)

class Ben19Dataset(Dataset):
    def __init__(self, lmdb_path, csv_path, img_transform='default'):
        """
        Parameter
        ---------
        lmdb_path:      path to the LMDB file for efficiently loading the patches.
        csv_path:       path to a csv file containing files that will make up the dataset
                        (e.g. train, ireland, val etc.)
        img_transform:  default does BEN19 specific interpolation and transformation
                        s.t. a PyTorch model can be trained.
        """
        self.lmdb_path = lmdb_path
        self.env = lmdb.open(lmdb_path, readonly=True, lock=False, readahead=False, meminit=False)
        self.env = None
        self.patch_names = self.read_csv(csv_path)
        self.transform = self.init_transform(img_transform)

    def read_csv(self, csv_data):
        patch_names = []
        with open(csv_data, 'r') as f:
            csv_reader = csv.reader(f)
            for row in csv_reader:
                patch_names.append(row[0])
        return patch_names

    def init_transform(self, img_transform):
        if img_transform == 'default':
            return transforms.Compose([ToTensorBEN19(),
                                       NormalizeBEN19()])
        else:
            return img_transform
    
    def _init_db(self):
        self.env = lmdb.open(self.lmdb_path, readonly=True, lock=False, readahead=True, meminit=False)
    def __getitem__(self, idx):
        """Get item at position idx of Dataset."""
        if self.env is None:
            self._init_db()
        patch_name = self.patch_names[idx]

        with self.env.begin(write=False) as txn:
            byteflow = txn.get(patch_name.encode())

        patch_data = loads_pyarrow(byteflow)
        
        
        
        
        s1_patch = patch_data.s1_patch
        s1_bands = np.float32(s1_patch.get_stacked_bands())
        

                                  
        label = ben_19_labels_to_multi_hot(patch_data.new_labels)
        
      
        
        sample = dict(s1_bands=s1_bands, label=label, index=idx)
        if self.transform is not None:
                            sample = self.transform(sample)

        return sample

    def __len__(self):
        """Get length of Dataset."""
        return len(self.patch_names)


class NormalizeBEN19:


    def __init__(self, normalization: str = "standard"):


        ben_means = BAND_STATS["mean"]
        ben_stds = BAND_STATS["std"]

        self.s1_bands_mean = torch.Tensor([-12.619993741972035, -19.29044597721542])
        self.s1_bands_std = torch.Tensor([5.115911777546365, 5.464428464912864])
        self.s1_bands_min = torch.Tensor([-74.33214569091797, -75.11137390136719])
        self.s1_bands_max = torch.Tensor([34.60696029663086, 33.59768295288086])

        if normalization == "standard":
            self.normalize_s1_bands = torchvision.transforms.Normalize(self.s1_bands_mean, self.s1_bands_std)
        elif normalization == "minmax":
            self.normalize_s1_bands = MinMaxNormalization(min_=self.s1_bands_min, max_=self.s1_bands_max)
        else:
            raise ValueError(f"Normalization type {normalization} is not supported for S1 data."
                             f"Please select one of the following none|standard|minmax.")

    def __call__(self, sample: dict):
        """Applies the specified normalisation operation to the S1 bands.
        :param sample: The dictionary containing entries for S2 10m bands, S2 20m bands, S1 bands and the labels.
        :return: The return dictionary with the same structure as sample."""
        s1_bands = sample['s1_bands']
        s1_bands = self.normalize_s1_bands(s1_bands)
        sample['s1_bands'] = s1_bands

        return sample


class ToTensorBEN19:
    """BEN specific conversion of ndarrays in sample to Tensors."""

    def __call__(self, sample: dict) -> dict:
 
        sample['s1_bands'] = torch.tensor(sample['s1_bands'])
        sample['label'] = torch.tensor(sample['label'])

        return sample
